/* Name: main.c
 * Author: James Fletcher
 * Copyright: (c) James Fletcher
 * License: This work may be reproduced, modified, distributed, performed, and displayed for any purpose, but must acknowledge "fab-synth". Copyright is retained and must be preserved. The work is provided as is; no warranty is provided, and users accept all liability.
 */


#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>


// IO pin usage
#define CV_SYNC_IN      _BV(PB2)

#define AUDIO_OUT_PIN   _BV(PA7) // Audio PWM out on OC0B (PA7)

#define SPI_DI          _BV(PA6)
#define SPI_DO          _BV(PA5)
#define SPI_SCK         _BV(PA4)

#define LED_PIN         _BV(PA3)

#define POT3            _BV(PA2)
#define POT2            _BV(PA1)
#define POT1            _BV(PA0)


/*
 * SPI frame format for audio and cv data
 * Total of 20 bytes timeslots are:
 *   sync, cv-hi, cv-mid, cv-lo, aud0, ... aud15
 * Sync timeslot is empty, it just contains a falling edge on the sync line signaling the start of a new frame
 */
#define FRAME_SYNC_SLOT     0
#define FRAME_CV0_SLOT      1
#define FRAME_CV1_SLOT      2
#define FRAME_CV2_SLOT      3
#define FRAME_AUDIO_START_SLOT   4
// We have 3 frame buffers:
// 1 - currently receiving data
// 2 - currently DSP proccessing data
// 3 - currently sending to the next node
uint8_t rxFrame[20];
uint8_t processingFrame[20];
uint8_t txFrame[21]; // we que up byte beyond end of buffer but never send, hence 1 more byte here
uint8_t rxFrameSlotIndex = 0; // index of time slot in frame
volatile uint8_t bufferReady = 0; // semaphore to main code that a new buffer is ready to process

// CV data
// cv0: bit 7 - trigger
// cv0: bit 6 - gate
// cv1: unused
// cv2: control voltage value
#define CV0_TRIG 0x20
#define CV0_GATE 0x40

// Values of pots, updated live
uint16_t pots[3];
#define POT_ATTACK  2
#define POT_SUSTAIN 0
#define POT_RELEASE 1

// Envelope phase (state machine)
uint8_t envelopePhase = 0;  // current phase of state machine
int32_t envelopeScale = 0; // scale for this sample as we sweep through envelope 16:16 fixed point
uint32_t envelopeCount = 0; // index of current sample from start of phase
#define ENV_NONE    0
#define ENV_ATTACK  1
#define ENV_DECAY   2
#define ENV_SUSTAIN 3
#define ENV_RELEASE 4


// Audio PWM
uint8_t audioPWMSlot = 0;

uint8_t waveIndex = 0; // used to build temporary waveforms




void cycleFrameBuffers(void);



// Update PWM output - old code
// OCR0B is double buffered and only updated as counter crosses bottom, so no need to update here
ISR(TIM0_COMPB_vect)
{
    // Update waveform output
    //OCR0B = triLookup[waveIndex++];
}


// Timer 1 compare/overflow - update audio PWM duty cycle on Timer 0 with next audio sample
ISR(TIM1_COMPA_vect)
{
    // Update waveform output
    audioPWMSlot++;
    audioPWMSlot &= 0x0f; // range limit for case when no clock module present
    OCR0B = txFrame[FRAME_AUDIO_START_SLOT+audioPWMSlot];
    
}


// USI counter overflow indicates SPI byte rx
ISR(USI_OVF_vect)
{
    // Clear irq flag
	USISR = _BV(USIOIF);
    
    // Move rx byte into buffer
    rxFrame[rxFrameSlotIndex] = USIDR;
    
    // Update index
    rxFrameSlotIndex++;
    
    // Send next tx byte
    USIDR = txFrame[rxFrameSlotIndex]; // may be byte 21 hence larger tx buffer

    if(rxFrameSlotIndex >= 20)
    {
        cycleFrameBuffers(); // Move frame somewhere for processing
        
        // Signal new frame arrived to DSP code
        bufferReady = 1;
        
        // debug
        //if(rxFrame[3] != 0)
        //    PORTA &= ~LED_PIN; // Some key, LED on.
        //else
        //    PORTA |= LED_PIN; // LED off
        rxFrameSlotIndex = 1; // byte 0 (sync) never sent
    }
}

// INT0 pin IRQ
// SYNC signal from clock unit
// Setup to generate IRQ on falling edge of sync pin
ISR(EXT_INT0_vect)
{
    // Re-sync audio smaple rate clock in Timer1
    // TCNT1 irq should be due to fire any cycle now, so reset counter as if irq had fired and put 1st byte on tx frame into DAC
    TCNT1 = 0;
    audioPWMSlot = 0;
    OCR0B = txFrame[FRAME_AUDIO_START_SLOT+audioPWMSlot];
    
    
    // Output 1st byte of our frame ready for transfer session
    USIDR = txFrame[1];

    // Reset to start of frame
    rxFrameSlotIndex = 1; // byte 0 (sync) never sent
    // Reset SPI rx to start of a new byte
    USISR = 0b11100000; // clear flags, zero counter
}


// Called each time we recieve a new frame from SPI
void cycleFrameBuffers(void)
{
    int i;
    
    // Move all frame buffers around
    // No need to move byte 0 (Sync) as it's empty
    for(i=1; i<20; i++)
    {
        txFrame[i] = processingFrame[i];
        processingFrame[i] = rxFrame[i];
    }
}


// Scale audio signal
// Range is 0 -> 256 as 0 -> 1
void scaleBuffer(uint8_t scale)
{
    uint8_t i;
    uint16_t temp;
    
    for(i=0; i<16; i++)
    {
        temp = (uint16_t)processingFrame[FRAME_AUDIO_START_SLOT+i] * (uint16_t)scale;
        temp = temp>>8;
        processingFrame[FRAME_AUDIO_START_SLOT+i] = (uint8_t)temp;
    }
}


// Calculate the envelope on the signal
void envelopeBufferORIG(void)
{
    uint8_t i;
    uint32_t temp;
    uint32_t attackRate;
    
    // Set up some rate variables
    // Attack rate 16:16 fixed point step to add per sample during attack phase
//    attackRate = (uint32_t)pots[POT_ATTACK] * 31250L;
//    attackRate /= 1024L;
//    attackRate = (attackRate == 0) ? (256L << 16) : (256L << 16) / attackRate; // div zero check
    
    // debug
//    attackRate = 500;
//    envelopeScale = 32767L; //30*1000; //500L*4L; //1L << 15;

    for(i=0; i<16; i++)
    {
        // Get sample
        temp = /*(uint32_t)*/processingFrame[FRAME_AUDIO_START_SLOT+i];
        
//        switch(envelopePhase)
//        {
//            case ENV_NONE: // just leave sample in buffer alone
//                break;
//            case ENV_ATTACK:
//                envelopeScale += attackRate;
//                temp = (uint8_t)((temp * envelopeScale) >> 16);
                temp = /*(uint8_t)*/((temp * 16383) >> 16);
//                if(envelopeScale >= (256L << 16))
//                {
//                    envelopeScale = 256L << 16;
//                    envelopePhase = ENV_SUSTAIN;
//                }
//                break;
//            case ENV_SUSTAIN:
//                break;
//            case ENV_RELEASE:
//                break;
//        }
        
        // Write sample
        processingFrame[FRAME_AUDIO_START_SLOT+i] = (uint8_t)temp;
    }
}

void envelopeBuffer(void)
{
    uint8_t i;
    uint16_t temp;
    uint32_t attackRate, releaseRate;
    int32_t sustainScale;
    uint8_t scale;
    
    // Set up some rate variables
    // Attack rate 16:16 fixed point step to add to scale per sample during attack phase
    attackRate = (uint32_t)pots[POT_ATTACK] * 31250L;
    attackRate /= 1024L;
    attackRate = (attackRate == 0) ? (256L << 16) : (256L << 16) / attackRate; // div zero check
    
    // Release rate 16:16 fixed point step to add to scale per sample during release phase
    releaseRate = (uint32_t)pots[POT_RELEASE] * 31250L;
    releaseRate /= 1024L;
    releaseRate = (releaseRate == 0) ? (256L << 16) : (256L << 16) / releaseRate; // div zero check
    
    // Sustain scale 16:16 fixed point with int part range 0 -> 256, scale at which we start to 'sustain' during the release period
    sustainScale = ((uint32_t)pots[POT_SUSTAIN]) << 14;
    if(sustainScale >= 256L<<16)
        sustainScale = 255L<<16;
    
    for(i=0; i<16; i++)
    {
        switch(envelopePhase)
        {
            case ENV_NONE: // just leave sample in buffer alone
                envelopeScale = 1L<<16;
                break;
            case ENV_ATTACK:
                envelopeScale += attackRate;
                if(envelopeScale >= (256L << 16))
                {
                    envelopeScale = 256L << 16;
                    envelopePhase = ENV_DECAY;
                }
                break;
            case ENV_DECAY:
                envelopeScale -= releaseRate;
                if(envelopeScale <= sustainScale)
                {
                    envelopeScale = sustainScale;
                    envelopePhase = ENV_SUSTAIN;
                }
                break;
            case ENV_SUSTAIN:
                // do nothing until we loose the 'gate' sgnal
                if((processingFrame[FRAME_CV0_SLOT] & CV0_GATE) == 0)
                    envelopePhase = ENV_RELEASE;
                break;
            case ENV_RELEASE:
                envelopeScale -= releaseRate;
                if(envelopeScale <= 0)
                {
                    envelopeScale = 0;
                    envelopePhase = ENV_NONE;
                }
                break;
        }
        
        // for some reason only the 16bit maths works, 32 bit fails!
        // Scale is a 8:8 fixed point
        scale = envelopeScale>>16;
        if(scale < 0) scale = 0;
        
        // Update sample
        temp = (uint16_t)processingFrame[FRAME_AUDIO_START_SLOT+i] * (uint16_t)scale;
        temp = temp >> 8;
        processingFrame[FRAME_AUDIO_START_SLOT+i] = (uint8_t)temp;
    }
}


int main(void)
{
    uint8_t i;
    uint8_t current_ADC_Chan = 0;
    uint8_t gate=0; // bodge
    
    for(i=0; i<20; i++)
        rxFrame[i] = processingFrame[i] = txFrame[i] = 0;
    
    for(i=0; i<3; i++)
        pots[i] = 0;
    
    // Setup outputs
    DDRA = LED_PIN | AUDIO_OUT_PIN | SPI_DO;
    //DDRB = CV_SYNC_IN;
    
    // Enable pull ups on SPI input pins?
    PORTA = SPI_DI | SPI_SCK;
    
    // Turn LED off
    PORTA |= LED_PIN; //&= ~LED_PIN;
    
    
    // TCNT0 Generates our fast 78KHz PWM signal for our DAC
    // Setup TCNT0 for PWM Fast mode
    // Timer cycles at 78125Hz
    TCCR0A = 0b00100011; // Clear OC0B on match, Fast PWM (mode 3)
    TCCR0B = 0b00000001; // Fast PWM, No prescale
    TCNT0 = 0;
    OCR0B = 128+64;
    //TIMSK0 = 0b00000100; // Enable OC0B match irq - no longer update OC0B here
    
    // TCNT1 Generates PWM duty cycle update events, ie. times out at the audio sample frequency of 31KHz
    // I had wanted to use SPI rx irq for this, but now we have our cv data in the audio stream there is no longer a 1:1 between SPI bytes and audio sample points
    // SPI bytes turn up at 39062.5Hz.  16 out of every 20 are audio samples giveing 31250Hz audio sample rate
    TCCR1A = 0b00000000; // No OC1x functions, CTC (Mode 4)
    TCCR1B = 0b00001001; // CTC, no pre-scaler
    TCNT1 = 0;
    OCR1A = 700; //F_CPU/31250; // 640 - gives 31250Hz
    TIMSK1 = 0b00000010; // enable compare A irq

    // Enable SPI
    USICR = 0b01011000; // irq on overflow, 3 wire mode, ext clock pos edge
    USISR = 0b11100000; // clear flags, zero counter
    
    // Enable INT0 (sync pin), falling edge generates irq
    MCUCR = (MCUCR & 0xfc) | 0x02;
    GIMSK |= _BV(INT0);
    
    sei(); // Global IRQ enable
    
    for(;;)
    {
        // Read the 3 pots and store the values
        // Force long ADC conversion as we're swapping channels.  Not sure it's really required for single ended inputs
        //ADCSRA = 0;
        // Set ADC channel
        ADMUX = 0b00000000 | (current_ADC_Chan& 0x1f); // Ref=VCC, channel = current
        ADCSRA = 0b10000100; // ADC enable, no irqs, prescale / 16 (need to slow ADC slightly as my pots are 50K and take a while to charge the SH unit)
        ADCSRA |= (1 << ADSC); // Initiate conversion
        while (ADCSRA & (1 << ADSC)) // Wait for it to complete
            ;
        pots[current_ADC_Chan] = 1024 - ADC; // Pots are soldered in back to front, so invert the range
        current_ADC_Chan++;
        if(current_ADC_Chan > 2) current_ADC_Chan = 0;

        // Check for new procesisng to be done
        if(bufferReady > 0)
        {
            bufferReady = 0;
            
            // Debug
            //if(processingFrame[FRAME_CV0_SLOT] != 0)
            //if((processingFrame[FRAME_CV0_SLOT] & CV0_TRIG) == 0)
            //    ; //PORTA |= LED_PIN; //&= ~LED_PIN;
            //else
            //    PORTA &= ~LED_PIN;


            // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            // Why don't I get the TRIG signal???
            // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            //if((processingFrame[FRAME_CV0_SLOT] & CV0_TRIG) != 0)
            if((processingFrame[FRAME_CV0_SLOT] & CV0_GATE) == 0)
            {
                gate = 0;
                PORTA |= LED_PIN;
            }
            else
            {
                if(gate == 0)
                {
                    gate = 1;
                PORTA &= ~LED_PIN;
                // New keypress so start a new envelope
                envelopePhase = ENV_ATTACK;
                envelopeScale = 0; // allow it to carry on from previous when keys swapped quickly?
                envelopeCount = 0;
                }
            }
            //scaleBuffer(128);
            envelopeBuffer();
        }

    }
    
    return 0;
}
